/* Copyright 2003 Heikki Orsila <heikki.orsila@ee.tut.fi>

TODO:
interactive mode: ls, command launch support, pwd, ...
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <ftw.h>
#include <glob.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include "unix-shell.h"

#include "uade-os.h"
#include "../osdep/strl.c"

struct songnode {
  struct songnode *next;
  char filename[PATH_MAX];
};

static int ftw_func(const char *file, const struct stat *sb, int flag);

static struct songnode *SongList = 0;

static void (*us_interrupt)(void) = 0;

/* catch CTRL-C and cause REBOOT */
static void us_sigint_handler(int signum) {
  signum = signum;
  if (us_interrupt) {
    us_interrupt();
  } else {
    fprintf(stderr, "uade: non-initialized sigint interrupt\n");
  }
}

static void us_setup(void (*interrupt)(void)) {
  us_interrupt = interrupt;
  /* create CTRL-C break for cmdline tool */
  uade_create_signalhandler(us_sigint_handler, UADE_SIGINT);
}

static void us_enqueue_song(char *file) {
  struct songnode *s = SongList;
  struct songnode *new;

  /* allow queuing of zero length string */
  if (uade_is_regular_file(file) || strlen(file) == 0) {
    new = malloc(sizeof(struct songnode));
    new->next = 0;
    strlcpy(new->filename, file, sizeof(new->filename));
    if (s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
    }
  } else {
    struct stat st;
    if (!stat(file, &st)) {
      if (S_ISDIR(st.st_mode)) {
	ftw(file, ftw_func, 20);
      }
    }
  }
}

static int ftw_func(const char *file, const struct stat *sb, int flag) {
  if (flag == FTW_F) {
    char name[PATH_MAX];
    strlcpy(name, file, sizeof(name));
    printf("enqueing: %s\n", name);
    us_enqueue_song(name);
  }
  return 0;
}

static int us_song_list_empty(void) {
  return SongList ? 0 : 1;
}

static void us_get_next(char *dst, int maxsize) {
  char playername[PATH_MAX];
  char tempname[PATH_MAX];
  struct songnode *t;
  if (!SongList)
    return;
  strlcpy(dst, SongList->filename, maxsize);
  t = SongList;
  SongList = SongList->next;
  free(t);
}

static int prompt_shown = 0;
static int cmd_number = 1;
static int cmd_buf_size = 0;
static char cmd_buf[4096];

/* strip leading whitespace from cmd_buf */
static void strip_whitespace(void) {
  int i;
  i = 0;
  while (i < cmd_buf_size) {
    if (cmd_buf[i] == '\n') {
      prompt_shown = 0;
    }
    if (!isspace((int) cmd_buf[i])) {
      break;
    }
    i++;
  }
  memmove(cmd_buf, &cmd_buf[i], cmd_buf_size - i);
  cmd_buf_size -= i;
}

static int get_rest_of_line(char *buf, int maxlen) {
  int i;
  for (i=0; i<cmd_buf_size; i++) {
    if (cmd_buf[i] == '\n') {
      cmd_buf[i] = 0;
      strlcpy(buf, cmd_buf, maxlen);
      memmove(cmd_buf, &cmd_buf[i+1], cmd_buf_size - (i + 1));
      cmd_buf_size -= i + 1;
      return 1;
    }
  }
  return 0;
}

/* get a word from cmd_buf */
static int pop_word(char *buf, int maxlen) {
  int i;
  strip_whitespace();
  for (i=0; i<cmd_buf_size; i++) {
    if (isspace((int) cmd_buf[i])) {
      cmd_buf[i] = 0;
      strlcpy(buf, cmd_buf, maxlen);
      memmove(cmd_buf, &cmd_buf[i+1], cmd_buf_size - (i + 1));
      cmd_buf_size -= i + 1;
      return 1;
    }
  }
  return 0;
}

static int us_interaction(struct uade_command *cmd, int wait_for) {
  struct pollfd ifd;
  int ret, maxlen;
  char buf[4096];
  int i;

  if (!prompt_shown) {
    printf("uade %d> ", cmd_number);
    fflush(stdout);
    prompt_shown = 1;
    cmd_number++;
  }

  cmd->ret = 0;
  memset(&ifd, 0, sizeof(ifd));

  ifd.fd = 0;
  ifd.events = POLLIN;
  ret = poll(&ifd, 1, wait_for ? -1 : 0);
  if (ret > 0) {
    if (ifd.revents & POLLIN) {
      maxlen = sizeof(cmd_buf) - cmd_buf_size;
      if (maxlen > 0) {
	ret = read(ifd.fd, &cmd_buf[cmd_buf_size], maxlen);
	if (ret < 0) {
	  if (errno != EINTR) {
	    goto no_interaction;
	  }
	} else if (ret == 0) {
	  goto no_interaction;
	} else {
	  cmd_buf_size += ret;
	}
      }
    }
  } else if (ret < 0) {
    if (errno != EINTR) {
      goto no_interaction;
    }
  }

  ret = pop_word(buf, sizeof(buf));
  if (!ret) {
    return 0;
  }

  if (!strcmp("next", buf) || !strcmp("n", buf)) {
    cmd->type = UADE_SONG_END;
    goto got_command;
  }

  if (!strcmp("m", buf)) {
    cmd->type = UADE_REBOOT;
    goto got_command;
  }

  i = 0;
  while (buf[i]) {
    if (!isdigit((int) buf[i])) {
      goto not_a_number;
    }
    i++;
  }
  cmd->type = UADE_SETSUBSONG;
  cmd->ret = malloc(sizeof(int));
  if (!cmd->ret) {
    goto no_interaction;
  }
  *((int *) cmd->ret) = (int) atoi(buf);
  goto got_command;

 not_a_number:

  if (!strcmp("play", buf)) {
    strip_whitespace();
    if (get_rest_of_line(buf, sizeof(buf))) {
      glob_t pglob;
#ifndef GLOB_TILDE
#define GLOB_TILDE 0
#endif
      if (!glob(buf, GLOB_NOCHECK | GLOB_TILDE, 0, &pglob)) {
	for (i=0; i< ((int) pglob.gl_pathc); i++) {
	  fprintf(stderr, "enqueuing: %s\n", pglob.gl_pathv[i]);
	  us_enqueue_song(pglob.gl_pathv[i]);
	}
	globfree(&pglob);
      }
    }
    prompt_shown = 0;
    goto no_command;
  }

  if (!strcmp("cd", buf)) {
    strip_whitespace();
    if (get_rest_of_line(buf, sizeof(buf))) {
      glob_t pglob;
      if (!glob(buf, GLOB_NOCHECK | GLOB_TILDE, 0, &pglob)) {
	for (i=0; i< ((int) pglob.gl_pathc); i++) {
	  if (chdir(pglob.gl_pathv[i])) {
	    perror("can not change to that directory");
	  } else {
	    printf("pwd = %s\n", pglob.gl_pathv[i]);
	    break;
	  }
	}
	globfree(&pglob);
      }
    }
    prompt_shown = 0;
    goto no_command;
  }

  /*
    if (!strcmp("ls", buf)) {
    char lsline[4096];
    strip_whitespace();
    get_rest_of_line(buf, sizeof(buf));
    snprintf(lsline, sizeof(lsline), "ls %s", buf);
    system(buf);
    prompt_shown = 0;
    goto no_command;
    }
  */

  if (!strcmp("flush", buf)) {
    char fname[PATH_MAX];
    while (!us_song_list_empty()) {
      us_get_next(fname, sizeof(fname));
    }
    prompt_shown = 0;
    goto no_command;
  }

  if (strlen(buf) == 1) {
    switch (buf[0]) {
    case 'f':
      cmd->type = UADE_TOGGLE_LED;
      goto got_command;
    case 'h':
      printf("<number>\tGo to subsong <number>\n");
      printf("cd <dir>\tSwitch to directory <dir>\n");
      printf("f\t\tToggle LED Filter\n");
      printf("flush\t\tFlush playlist\n");
      printf("h\t\tDisplay help\n");
      printf("m\t\tNext song\n");
      printf("n\t\tNext (sub)song\n");
      printf("play <pattern>\tEnqueue songs matching shell pattern\n");
      prompt_shown = 0;
      goto no_command;
    default:
      break;
    }
  }

  printf("unknown command for interaction\n");
  prompt_shown = 0;

 no_command:
  return 0;

 got_command:
  prompt_shown = 0;
  return 1;

 no_interaction:
  cmd->type = UADE_NO_INTERACTION;
  return 1;
}

void us_functions(struct uade_local *local) {
  local->setup = us_setup;
  local->enqueue_song = us_enqueue_song;
  local->get_next = us_get_next;
  local->interaction = us_interaction;
  local->song_list_empty = us_song_list_empty;
};
